// Purpose.  Flyweight design pattern
//
// 1. Identify shareable state (intrinsic) and non-shareable state (extrinsic)
// 2. Create a Factory that can return an existing object or a new object
// 3. The client must use the Factory instead of "new" to request objects
// 4. The client (or a third party) must compute the extrinsic state

import java.awt.*;
import java.awt.event.*;


//-------------------------------------------------------------
//begin implementing chain of responsibility(processChain)
//-------------------------------------------------------------
interface processChain
{
    void process(int Number);
    void addLast(processChain nextP);
}



class primeP implements processChain{
   // private Handler; 
    private processChain nextP;
    private Handler primeH;
    
    public primeP(){primeH = new PrimeHandler();}
    public void process(int Number){
        //determine if the Number is prime
        boolean isPrime = true;
        if(Number <= 1) isPrime = false;
        else if(Number > 3)
        {
            for(int i = 2; i * i <= Number; i++)
            {
                if(Number % i == 0) {
                    isPrime = false;
                    break;
                }
            }
        }
        
        if(isPrime) primeH.calculate(Number);// client.receivedata(Number);
        else nextP.process(Number);
    }
     
    public void addLast(processChain nextP)
    {
        if( this.nextP == null) this.nextP = nextP;
        else this.nextP.addLast(nextP);
    }
}


class oddP implements processChain{
    private processChain nextP;
    private Handler oddH;
    
    public oddP(){ oddH = new OddHandler();}
    public void process(int Number){
        //determine if the Number is prime
        
        if(Number % 2 != 0) oddH.calculate(Number);
        else nextP.process(Number);
    }
       
    public void addLast(processChain nextP)
    {
        if( this.nextP == null) this.nextP = nextP;
        else this.nextP.addLast(nextP);
    }
}

class evenP implements processChain{
    private Handler evenH;
    private processChain nextP = null;
    public evenP(){ evenH = new EvenHandler(); }
    public void process(int Number){
        //determine if the Number is prime
        evenH.calculate(Number);
    }
       
    public void addLast(processChain nextP){}

}


//-------------------------------------------------------------
//begin implementing GUI using flyweight
//-------------------------------------------------------------
class FlyweightFactory {
   private static java.util.Hashtable ht = new java.util.Hashtable();
   private static ButtonListener bl = new ButtonListener();
   public static Button makeButton( String num ) {
      if (ht.containsValue( num ))
         return (Button) ht.get( num );        // 2. Return an existing object
      Button btn = new Button( num );          // 1. Identify intrinsic state
      btn.addActionListener( bl );
      ht.put( num, btn );
      return btn;                              // 2. Return a new object
   }
   public static void report() {
      System.out.print( "size=" + ht.size() + "   " );
      for (java.util.Enumeration e = ht.keys(); e.hasMoreElements(); )
         System.out.print( e.nextElement() + " " );
      System.out.println();
   }  
}

class ButtonListener implements ActionListener {
    
   
   public void actionPerformed( ActionEvent e) {
      Button      btn  = (Button) e.getSource();
      Component[] btns = btn.getParent().getComponents();
      float stuff[] = new float[500];
      int i = 0;
      for ( ; i < btns.length; i++)
         if (btn == btns[i]) break;
      // 4. A third party must compute the extrinsic state (x and y)
      //    (the Button label is intrinsic state)
      System.out.println( "label-" + e.getActionCommand()
         + "  x-" + i/Flyweight_assignment.NUM   + "  y-" + i%Flyweight_assignment.NUM );  // 1. Identify extrinsic state
      Flyweight_assignment.primeChain.process(Integer.parseInt(e.getActionCommand()));
   }  
}


class Flyweight_assignment {
   public static final int NUM = 15;
   public static final int RAN = 224;
   
   //client 1, 2, 3
   //public static processChain primeChain = new primeP(client1);
   
   
   
   public static processChain primeChain = new primeP();
   public static processChain oddChain = new oddP();
   public static processChain evenChain = new evenP();
   
   public static void main( String[] args ) {
      Frame frame = new Frame( "Flyweight Demo" );
      frame.setLayout( new GridLayout( NUM, NUM ) );
      for (int i=0; i < NUM; i++)
         for (int j=0; j < NUM; j++)
            // 3. The client must use the Factory to request objects
            frame.add( FlyweightFactory.makeButton( 
               Integer.toString( (int)(Math.random()*RAN) ) ) );
      frame.pack();
      frame.setVisible( true );
      FlyweightFactory.report();
      primeChain.addLast(oddChain);
      primeChain.addLast(evenChain);
   }  
}

// size=25   24 23 22 21 20 19 18 17 16 9 15 8 14 13 7 12 6 5 11 10 4 3 2 1 0
// label-23  x-0  y-0
// label-7  x-0  y-1
// label-21  x-1  y-1
// label-21  x-4  y-6
// label-7  x-9  y-9